"""
ABCI TCP server
Tendermint connects to this server over 3 different connections:
 - mempool: used for check_tx
 - consensus: used for the begin_block -> deliver_tx -> end_block -> commit flow
 - query: used to query application state

This can be a bit confusing in the app since gevent spawns a greenlet for each
connection.  If one crashes you will not have full connectivity to Tendermint
"""

import signal
from io import BytesIO
import asyncio
from contextlib import suppress


from .encoding import read_messages, write_message
from .utils import get_logger
from .application import BaseApplication

from .types_pb2 import (
    Request, Response, ResponseEcho,
    ResponseFlush, ResponseException
)

log = get_logger()


class ProtocolHandler:
    """ Internal handler called by the server to process requests from
    Tendermint.  The handler delegates call to your application"""

    def __init__(self, app):
        self.app = app

    def process(self, req_type, req):
        handler = getattr(self, req_type, self.no_match)
        return handler(req)

    def echo(self, req):
        msg = req.echo.message
        response = Response(echo=ResponseEcho(message=msg))
        return write_message(response)

    def flush(self, req):
        response = Response(flush=ResponseFlush())
        return write_message(response)

    def info(self, req):
        result = self.app.info(req.info)
        response = Response(info=result)
        return write_message(response)

    def set_option(self, req):
        result = self.app.set_option(req.set_option)
        response = Response(set_option=result)
        return write_message(response)

    def check_tx(self, req):
        result = self.app.check_tx(req.check_tx.tx)
        response = Response(check_tx=result)
        return write_message(response)

    def deliver_tx(self, req):
        result = self.app.deliver_tx(req.deliver_tx.tx)
        response = Response(deliver_tx=result)
        return write_message(response)

    def query(self, req):
        result = self.app.query(req.query)
        response = Response(query=result)
        return write_message(response)

    def commit(self, req):
        result = self.app.commit()
        response = Response(commit=result)
        return write_message(response)

    def begin_block(self, req):
        result = self.app.begin_block(req.begin_block)
        response = Response(begin_block=result)
        return write_message(response)

    def end_block(self, req):
        result = self.app.end_block(req.end_block)
        response = Response(end_block=result)
        return write_message(response)

    def init_chain(self, req):
        result = self.app.init_chain(req.init_chain)
        response = Response(init_chain=result)
        return write_message(response)

    def no_match(self, req):
        response = Response(exception=ResponseException(error="ABCI request not found"))
        return write_message(response)


def handler(signum, frame):
    raise Exception

signal.signal(signal.SIGQUIT, handler)
signal.signal(signal.SIGTERM, handler)
signal.signal(signal.SIGINT, handler)


class ABCIServer:
    def __init__(self, port=26658, app=None):
        if not app or not isinstance(app, BaseApplication):
            log.error("Application missing or not an instance of Base Application")
            raise TypeError("Application missing or not an instance of Base Application")
        self.port = port
        self.protocol = ProtocolHandler(app)

        self.loop = asyncio.get_event_loop()
        coro = asyncio.start_server(self.__handle_connection, '0.0.0.0', 26658, loop=self.loop)
        self.server = self.loop.run_until_complete(coro)

    def run(self):
        """Option to calling manually calling start()/stop(). This will start
        the server and watch for signals to stop the server"""
        log.info(" ABCIServer started on port: {}".format(self.port))
        try:
            self.loop.run_forever()
        except Exception:
            pass

        pending = asyncio.Task.all_tasks()
        list(map(lambda task: task.cancel(), pending))
        with suppress(asyncio.CancelledError):
            self.loop.run_until_complete(asyncio.gather(*pending))

        # Close the server
        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())
        log.info("Shutting down server")
        self.loop.close()


    # TM will spawn off 3 connections: mempool, consensus, query
    # If an error happens in 1 it still leaves the others open which
    # means you don't have all the connections available to TM
    async def __handle_connection(self, reader, writer):
        addr = writer.get_extra_info('peername')
        log.info(f' ... Connection from Tendermint: {addr[0]}:{addr[1]} ...')
        data = BytesIO()
        last_pos = 0

        while True:

            # Create a new buffer every time there is the possibility.
            # This avoids having a never ending buffer.
            if last_pos == data.tell():
                data = BytesIO()
                last_pos = 0

            inbound = await reader.read(1024 * 8)
            data.write(inbound)

            if not len(inbound):
                break

            # Before reading the messages from the buffer, position the
            # cursor at the end of the last read message.
            data.seek(last_pos)
            messages = read_messages(data, Request)

            for message in messages:
                req_type = message.WhichOneof('value')
                response = self.protocol.process(req_type, message)
                writer.write(response)
                await writer.drain()
                last_pos = data.tell()

        writer.close()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protobuf/github.com/tendermint/tendermint/crypto/merkle/merkle.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from protobuf.github.com.gogo.protobuf.gogoproto import gogo_pb2 as protobuf_dot_github_dot_com_dot_gogo_dot_protobuf_dot_gogoproto_dot_gogo__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='protobuf/github.com/tendermint/tendermint/crypto/merkle/merkle.proto',
  package='merkle',
  syntax='proto3',
  serialized_pb=_b('\nDprotobuf/github.com/tendermint/tendermint/crypto/merkle/merkle.proto\x12\x06merkle\x1a\x36protobuf/github.com/gogo/protobuf/gogoproto/gogo.proto\"2\n\x07ProofOp\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\"+\n\x05Proof\x12\"\n\x03ops\x18\x01 \x03(\x0b\x32\x0f.merkle.ProofOpB\x04\xc8\xde\x1f\x00\x42\x14\xc8\xe2\x1e\x01\xd0\xe2\x1e\x01\xe0\xe2\x1e\x01\xf8\xe1\x1e\x01\xa8\xe2\x1e\x01\x62\x06proto3')
  ,
  dependencies=[protobuf_dot_github_dot_com_dot_gogo_dot_protobuf_dot_gogoproto_dot_gogo__pb2.DESCRIPTOR,])




_PROOFOP = _descriptor.Descriptor(
  name='ProofOp',
  full_name='merkle.ProofOp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='merkle.ProofOp.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='key', full_name='merkle.ProofOp.key', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='merkle.ProofOp.data', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=136,
  serialized_end=186,
)


_PROOF = _descriptor.Descriptor(
  name='Proof',
  full_name='merkle.Proof',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ops', full_name='merkle.Proof.ops', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\310\336\037\000')), file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=188,
  serialized_end=231,
)

_PROOF.fields_by_name['ops'].message_type = _PROOFOP
DESCRIPTOR.message_types_by_name['ProofOp'] = _PROOFOP
DESCRIPTOR.message_types_by_name['Proof'] = _PROOF
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ProofOp = _reflection.GeneratedProtocolMessageType('ProofOp', (_message.Message,), dict(
  DESCRIPTOR = _PROOFOP,
  __module__ = 'protobuf.github.com.tendermint.tendermint.crypto.merkle.merkle_pb2'
  # @@protoc_insertion_point(class_scope:merkle.ProofOp)
  ))
_sym_db.RegisterMessage(ProofOp)

Proof = _reflection.GeneratedProtocolMessageType('Proof', (_message.Message,), dict(
  DESCRIPTOR = _PROOF,
  __module__ = 'protobuf.github.com.tendermint.tendermint.crypto.merkle.merkle_pb2'
  # @@protoc_insertion_point(class_scope:merkle.Proof)
  ))
_sym_db.RegisterMessage(Proof)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\310\342\036\001\320\342\036\001\340\342\036\001\370\341\036\001\250\342\036\001'))
_PROOF.fields_by_name['ops'].has_options = True
_PROOF.fields_by_name['ops']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\310\336\037\000'))
# @@protoc_insertion_point(module_scope)



from io import BytesIO


def encode_varint(number):
    # Shift to int64
    number = number << 1
    buf = b''
    while True:
        towrite = number & 0x7f
        number >>= 7
        if number:
            buf += bytes((towrite | 0x80,))
        else:
            buf += bytes((towrite,))
            break
    return buf

def decode_varint(stream):
    shift = 0
    result = 0
    while True:
        i = _read_one(stream)
        result |= (i & 0x7f) << shift
        shift += 7
        if not (i & 0x80):
            break
    return result

def _read_one(stream):
    c = stream.read(1)
    if c == b'':
        raise EOFError("Unexpected EOF while reading bytes")
    return ord(c)

def write_message(message):
    buffer = BytesIO(b'')
    bz = message.SerializeToString()
    buffer.write(encode_varint(len(bz)))
    buffer.write(bz)
    return buffer.getvalue()


def read_messages(reader, message):
    """Return an interator of the messages found in
    the `reader` (BytesIO instance)."""

    while True:
        try:
            length = decode_varint(reader) >> 1
        except EOFError:
            return

        data = reader.read(length)

        if len(data) < length:
            return

        m = message()
        m.ParseFromString(data)

        yield m
